clc
clear
close all

load('filter.mat')

%Calculated impedance Rp=47kOhm, Cp=47pF
Y0 = 1/47e3 + 1i*2*pi*10e6*47e-12;
Z0 = 1/Y0;


%Normalised frequency
wn=0.0035*pi;

Z_FIR=[];
Z_DFT=[];
Z_DFT_G=[];

noiseEval=false;
if noiseEval
    K=10*ones(1,1e2);
else
    %K=linspace(1,10,1e4);
    K=10;
end

for k=K

    N=round(k*2*pi/wn);

    %Create independent variable n
    n=0:N + length(filt.Coefficients);

    %Create noisy deformed signal
    u=sin(wn*n) + 0.1*sin(2*wn*n) + 0.05*sin(3*wn*n) + 0.05*(rand(1,length(n))-0.5);
    i=calculateCurrent(u,wn,10e6);

    %Apply FIR filter
    U=filter(filt,u);
    I=filter(filt,i);

    %Remove transient (work with last 2048 samples)
    u=cropVariable(u,N);
    i=cropVariable(i,N);
    U=cropVariable(U,N);
    I=cropVariable(I,N);
    n=0:N-1;

    % plotUI(n,u,i);
    % plotUI(n,U,I);
    
    Z_FIR(end+1)=(rms(U)/rms(I))*exp(1i*deg2rad(calculateAngle(U,I,k)));

    %Evaluation by DFT
    %y=cos(n*wn)+1i*sin(n*wn);
    y=exp(1i*2*pi*round(k)*n/N);
    U_dft=u*y';%xcorr(u,y,0);
    I_dft=i*y';%xcorr(i,y,0);
    Z_DFT(end+1)=(U_dft/I_dft);
%     Au=abs(U_dft)*2/N;
%     Ai=abs(I_dft)*2/N;
%     phi=calculateAngle(U,I,k);
    % fprintf('Vyhodnocení impedance Rp=47k\x03a9, Cp=47pF\n');
    % fprintf('Výpočet:\tZ=%f, \x03c6=%f°\n',abs(Z0),rad2deg(angle(Z0)));
    % fprintf('FIR.\t\tZ=%f, \x03c6=%f°\n',rms(U)/rms(I),calculateAngle(U,I,k));
    % fprintf('DFT_def:\tZ=%f, \x03c6=%f°\n',Au/Ai,rad2deg(angle(U_dft)-angle(I_dft)));


    %Goertzel algorythm
    % U_g=goertzel(u,6);
    % I_g=goertzel(i,6);
    U_g=myGoertzel(u,round(k));
    I_g=myGoertzel(i,round(k));
    Z_DFT_G(end+1)=(U_g/I_g);
    % [Q,I]=goertzelSTM(u,5,length(u));
    % U_g = I + 1i*Q;
    % [Q,I]=goertzelSTM(i,5,length(i));
    % I_g = I + 1i*Q;
    % Au=abs(U_g)*2/N;
    % Ai=abs(I_g)*2/N;
    % phi=rad2deg(angle(U_g)-angle(I_g));
    % if phi>180
    %     phi=phi-360;
    % end
    % fprintf('DFT_goer:\tZ=%f, \x03c6=%f°\n',Au/Ai,phi);
end

d_A_FIR=100*(abs(Z_FIR)-abs(Z0))/abs(Z0);
d_A_DFT=100*(abs(Z_DFT)-abs(Z0))/abs(Z0);
d_A_DFT_G=100*(abs(Z_DFT_G)-abs(Z0))/abs(Z0);

d_PH_FIR=100*(angle(Z_FIR)-angle(Z0))/angle(Z0);
d_PH_DFT=100*(angle(Z_DFT)-angle(Z0))/angle(Z0);
d_PH_DFT_G=100*(angle(Z_DFT_G)-angle(Z0))/angle(Z0);

if noiseEval
    K=1:100;
end
figure();set(gcf,'position',[10,10,800,600])
subplot(2,1,1);plot(K,[d_A_FIR ; d_A_DFT ; d_A_DFT_G])
title('Relativní odchylka velikosti')
ylabel('Relativní odchylka [%]');
%xlabel('Násobek periody [-]');
legend('RMS po filtraci','DFT dle definice','Goertzelův algoritmus')
box on
grid on

%figure();set(gcf,'position',[10,10,800,600])
subplot(2,1,2);plot(K,[d_PH_FIR ; d_PH_DFT ; d_PH_DFT_G])
title('Relativní odchylka fáze')
ylabel('Relativní odchylka [%]');
if noiseEval
    xlabel('Číslo simulace [-]');
else
    xlabel('Násobek periody [-]');
end
legend('V časové oblasti','DFT dle definice','Goertzelův algoritmus','location','southeast')
box on
grid on

function y=cropVariable(x,N)
    y=x(length(x)-N+1:end);
end

function i=calculateCurrent(u,wn,f)
    %Parallel capacity 47pF
    C=47e-12;
    %Parallel resistance 47kOhm
    R=47e3;
    
    dt= wn/(2*pi*f);
    
    i=zeros(1,length(u));    
    u=[0 u];
    for x=2:length(u)-1
        i(x) = u(x)/R + C*(u(x)-u(x-1))/dt;
    end
end

function phi=calculateAngle(u,i,k)
    [uZeros_rising uZeros_iPos]=getZeros(u);
    [iZeros_rising iZeros_iPos]=getZeros(i);
    
    n=length(uZeros_rising);
    if(mod(n,2)~=0)
        n=n-1;
    end
    
    phi=[];
    for i=1:2:k-1;
        if( (uZeros_rising(i)==iZeros_rising(i)) )
            phi(end+1)=360*( iZeros_iPos(i)-uZeros_iPos(i) + iZeros_iPos(i+1)-uZeros_iPos(i+1) ) / (2*(uZeros_iPos(i+2)-uZeros_iPos(i)));
        else
            phi(end+1)=inf;
        end
    end
    phi=mean(phi);
end

function [rising,iPos]=getZeros(x)
    rising=[];
    iPos=[];
    for i=2:length(x)
        if ( x(i-1) < 0 ) && ( x(i) > 0)
            rising(end+1)=true;
            dy = x(i)-x(i-1);
            iPos(end+1)= i-1 - x(i-1)/dy;            
        end
        if ( x(i-1) > 0 ) && ( x(i) < 0)
            rising(end+1)=false;
            dy = x(i)-x(i-1);
            iPos(end+1)= i-1 - x(i-1)/dy;
        end
    end
end

function y=myGoertzel(x,k)
    N=length(x);
    x=single(x);
    Ak=single(2*cos(2*pi*k/N));
    WN=single(exp(-1i*2*pi*k/N));
    s=zeros(1,3);
    for n=1:N-1
        s(1)=x(n) + Ak*s(2) - s(3);
        s(3)=s(2);
        s(2)=s(1);
    end
    s(1)=x(n) + Ak*s(2) - s(3);
    y=s(1) - WN*s(2);
end
